local config_name = "MDI TargetList"
local defaults = {
	fontscale = {val = 1, Lval = 0, Uval = 2, dp = 2, name = "Font Scale"},
	target_color = {val = "65 255 65 65 *", Lval = 0, Uval = 255, name = "Highlight"},
	list_count = {val = 10, Lval = 1, name = "List Count"},
	name_width = {val = 0, Lval = 0, name = "Name Width"},
	x = {val = -1, Lval = 0, Uval = gkinterface.GetXResolution(), name = "x"},
	y = {val = -1, Lval = 0, Uval = gkinterface.GetYResolution(), name = "y"},
	hide_turrets = {val = 1, name = "Hide Turrets"},
	show_npc_ships = {val = 1, name = "Show NPC Ships"},
	show_all_players = {val = 1, name = "Show All Players"},
}

declare("targetlist", {
	version               = "3.2.1",
	config_name           = config_name,
	hide_turrets          = gkini.ReadInt   (config_name, defaults.hide_turrets.name, defaults.hide_turrets.val) == 1,
	invert_faction_filter = gkini.ReadInt   (config_name, "Invert Faction Filter", 0) == 1,
	filter                = gkini.ReadInt   (config_name, "Filter", 0), -- .filters will hold the IUP structure
	faction_filter        = gkini.ReadInt   (config_name, "Faction Filter", 0),
	fontscale             = gkini.ReadString(config_name, defaults.fontscale.name,        defaults.fontscale.val),
	current_target_color  = gkini.ReadString(config_name, defaults.target_color.name,     defaults.target_color.val),
	list_count            = gkini.ReadInt   (config_name, defaults.list_count.name,       defaults.list_count.val),
	name_width            = gkini.ReadInt   (config_name, defaults.name_width.name,       defaults.name_width.val),
	x                     = gkini.ReadInt   (config_name, defaults.x.name,                defaults.x.val),
	y                     = gkini.ReadInt   (config_name, defaults.y.name,                defaults.y.val),
	show_npc_ships        = gkini.ReadInt   (config_name, defaults.show_npc_ships.name,   defaults.show_npc_ships.val) == 1,
	show_all_players      = gkini.ReadInt   (config_name, defaults.show_all_players.name, defaults.show_all_players.val) == 1,
	default_target_color  = defaults.target_color.val,
	list_offset           = 1,
	playerlist            = {},
	shiplist              = {},
	id                    = "targetlist", -- used in MDI
	refresh_timer         = Timer(),
	refresh_delay         = 1000,
	current_target_listindex,
	current_target_nodeid,
	current_target_objectid,
	defaults = defaults,
})
if (targetlist.x == -1) then targetlist.x = nil end
if (targetlist.y == -1) then targetlist.y = nil end
if (targetlist.name_width == 0) then targetlist.name_width = nil end

console_print("MDI-TargetList loaded (v"..targetlist.version..").")

local function IsDeclared(value) return pcall(loadstring("return "..value)) end
	
local function IsIUP(control) return control and pcall(iup.GetType, control) end

local function loadluafile(...)
	local f, err = loadfile((...))
	assert(f and type(f) == "function", err)
	return f(select(2, ...))
end

local function check_number(num, Lval, Uval, dp)
	-- verify a number is within the given range
	-- <num> the number to check
	-- <Lval>, <Uval>: Lower and Upper bounds of the number range; if missing, compares the number against itself
	-- [dp]:  how many decimal places to consider
	num = tonumber(num)
	Lval = tonumber(Lval) or num
	Uval = tonumber(Uval) or num
	if (num and Lval <= num and Uval >= num) then
		if (dp) then
			local decimals = tostring(num):match("%.(%d+)") or "0"
			if (decimals and decimals:len() <= dp) then
				return num -- entered number has no more decimal places than specified
			else
				return false -- too many decimal places
			end
		else
			return num -- non-decimal number is within range
		end
	end
	return false  -- you done fucked up
end

function targetlist:SaveSettings()
	gkini.WriteInt   (config_name, defaults.hide_turrets.name,     self.hide_turrets and 1 or 0)
	gkini.WriteString(config_name, defaults.fontscale.name,        self.fontscale)
	gkini.WriteString(config_name, defaults.target_color.name,     self.current_target_color)
	gkini.WriteInt   (config_name, defaults.list_count.name,       self.list_count or defaults.list_count.val)
	gkini.WriteInt   (config_name, defaults.name_width.name,       self.name_width or defaults.name_width.val)
	gkini.WriteInt   (config_name, defaults.x.name,                self.x or defaults.x.val)
	gkini.WriteInt   (config_name, defaults.y.name,                self.y or defaults.y.val)
	gkini.WriteInt   (config_name, defaults.show_npc_ships.name,   self.show_npc_ships and 1 or 0)
	gkini.WriteInt   (config_name, defaults.show_all_players.name, self.show_all_players and 1 or 0)
end

local capship_type_names = {
	["Capella"] = true,
	["Goliath"] = true,
	["Heavy Assault Cruiser"] = true,
    ["Leviathan"] = true,
	["TPG Teradon Frigate"] = true,
    ["Trident Light Frigate"] = true,
    ["TPG Constellation Heavy Transport"] = true
}

local function filter_players(factionid)
	if (targetlist.faction_filter == 0) then
		return true
	elseif (not targetlist.invert_faction_filter and factionid == targetlist.faction_filter) then
		return true
	elseif (targetlist.invert_faction_filter and factionid ~= targetlist.faction_filter) then
		return true
	end
	return false
end

local function scan_player(charid)
	if (not charid) then return end
	local distance = GetRadarDistance(charid)
	local insert_player = false
	if (distance or targetlist.show_all_players) then
		local name = GetPlayerName(charid)
		local ship_name = GetPrimaryShipNameOfPlayer(charid)
		local is_npc = name:byte() == 42 -- 42 = *; byte is faster than match
		local factionid = GetPlayerFaction(charid)
		if     (targetlist.hide_turrets and name:match("\*.*Turret")) then
		    --do nothing

		elseif (targetlist.filter == 0) then -- All
			insert_player = filter_players(factionid)

		elseif (targetlist.filter == 1 and is_npc) then -- NPCs
			insert_player = filter_players(factionid)

		elseif (targetlist.filter == 2 and not is_npc)then -- PCs
			insert_player = filter_players(factionid)

		elseif (targetlist.filter == 3) then -- Capships
			if (filter_players(factionid)) then
				if (capship_type_names[ship_name] or (ship_name == "Robot" and name:match("Queen"))) then
					insert_player = true
				end
			end
		end

		if (is_npc and targetlist.show_npc_ships) then
			name = ship_name ~= "Robot" and "*"..(ship_name or "Unknown") or name
		end

		if (insert_player) then
			table.insert(targetlist.shiplist, {
				charid = charid,
				factionid = factionid,
				name = name,
				distance = math.floor(distance or "9999")
			})
		end
	end
	
	return scan_player(table.remove(targetlist.playerlist))
end

local function scan_sector()
	targetlist.refresh_timer:Kill()
	targetlist.playerlist = {}
	targetlist.shiplist = {}

	ForEachPlayer(function(charid)
		if (charid ~= GetCharacterID()) then table.insert(targetlist.playerlist, charid) end
	end)
	scan_player(table.remove(targetlist.playerlist)) -- initiates a recursive tail call to iterate over the entire player list
	table.sort(targetlist.shiplist, function(a, b) return a.distance < b.distance end)
	targetlist.display_container.update_display()

	targetlist.refresh_timer:SetTimeout(targetlist.refresh_delay, scan_sector)
end

loadluafile("display.lua", targetlist, scan_sector)
local change_target = loadluafile("commands.lua", targetlist, check_number)
loadluafile("events.lua", targetlist, IsDeclared, IsIUP, scan_sector)
loadluafile("touch.lua", targetlist, change_target)
loadluafile("help.lua", targetlist)
loadluafile("options.lua", targetlist, check_number)
